/*====================================================================

 - Integrated Desktop for TI89/TI92+/V200 -
 by Olivier Armand (ExtendeD)
 <ola.e-ml@wanadoo.fr>

 current version : 0.2c
 Initial Date : 10/09/2002
 Last Update  : 4/27/2003
 See ReadMe.txt for details.

 Your are having a look at this source code because your are curious
 about the functioning of ID. Of course I like your curiosity,
 that's why ID is an open-source project.

 But please ask me before using any part of this source code in your
 programs.
 
 If you have any questions about this source code, e-mail me.

===================================================================

 Compile with the GCC Switches :
 -fno-function-cse -mregparm=6 -fomit-frame-pointer
 But you should use the 'build' batch file to avoid absolute
 addressing modes, else ID will not work !

 Some hints to correct the invalid addressing modes that may appear
 in the .S file :

 replace
 	move.l #.LXXX,(%sp)
 by
 	pea .LXXX

 replace
 	move.l #HandlingEvent,20(%sp)
 	move.l 20(%sp),%a0
 by
 	lea HandlingEvent,%a0
 	lea 20(%sp),%a1
 	move.l %a0,(%a1)

 replace
 	add.l #Shortcuts,%d0
 	move.l %d0,-(%sp)
 by
 	pea Shortcuts
 	add.l %d0,(%a7)

===================================================================*/


#define USE_TI89
#define USE_TI92PLUS
#define USE_V200

// The following switch may not work if your version of TIGCC is too
// old. Remove it and replace it by #define MIN_AMS 206 and #define NO_AMS_CHECK
// (but there will not be any AMS check at the beginning of the program)
#define MIN_AMS 208     // We need the Flash Apps support and the Desktop
// Tipatch.lib must *not* create the variable '__calculator'. There is a
// calculator check in the TSR, and the variable '__calulator' wouldn't be in
// the TSR block. So we use an inline-detection :
#define NO_CALC_DETECT
#define USE_FLINE_ROM_CALLS
#define NO_EXIT_SUPPORT

#include <tigcclib.h>

#if (defined(__TIGCC__VERSION) && (__TIGCC_VERSION__ < 94))
#error You need TIGCC v0.94 or higher to build this project
#endif
// (http://tigcc.ticalc.org)

#include "TIFSimport.h"
#include "macros.h"
#include "structures.h"
#include "functions.h"
#include "install_sub.h"
#include "RD_Icon.h"
#include "tsr.h"
#include "h220xtsr.h"


//--------------------------------------------------------------------------

// the limits of the event hook (defined with inline assembly in tsr.h)
extern char NewEvHook;
extern char EndNewEvHook;
// the 4 bytes where the old event hook vector is stored in the tsr
extern EVENT_HANDLER OldEvHook;

//--------------------------------------------------------------------------


void _main(void)
{
	HANDLE ICO_handle; // handle of the data file containing the icons & the descriptions
	HANDLE FilesList_handle; // the handle of the list of the name of all the supported files
	unsigned short FilesNumber; // the number of files in this list
	char *FilesListPtr_top; // a pointer to this list
	char *NewEvHookPtr; // pointer to the block where the new event hook is installed
	BOOL TextViewerFound; // true if the text viewer was found 
	AppID RefreshDT_AppID = H_NULL; // the AppID of the fake app 'Refresh Desktop'
																  // (initialization to avoid a warning)
	
	// get a pointer to the first folder of the Desktop Folders Structure
	// DTfolders_Ptr is defined in tsr.h (it will be install with the event
	// hook that uses it)
	// This pointer will also be used if we are uninstalling ID (by
	// RemoveFromFolders() ) and if we restore the backup of the folders structure.
	if (!(DTfolders_Ptr = GetDT_FoldersPtr ()))
		// can't find the folders of the Desktop
		fatal ("Error: can't find the Desktop folders.");
	

	// test if the EV_hook is already installed
	if(EV_hook) {
			char *HookPtr = (char*)EV_hook;
			char *PrevHookPtr = NULL; // points to the previous hook in the linked list
			
			while (HookPtr != NULL)
			{
				HookPtr -= 16;     // point to the header of the hook
				// is the hook following Kevin Kofler's convention ?
				if(*(long*)HookPtr != EVHK_TAG_HEXA)
					break;     // no -> our hook can't be further in the linked list
				
				if(!memcmp(HookPtr, EV_HOOK_NAME, 12)) {
					// Already installed : uninstall it.
					// remove the fake apps of all the Desktop folders
					// (including 'All' ! because if ID is run again to install it, 
					// GetDT_FoldersPtr() will read the list of the files in the folder
					// 'All' which might be invalid if we do not clean it up)
					RemoveFromFolders (FALSE, H_NULL);
					// remove all the ACBs and object frames created
					FreeCreatedACB_and_Frames ();
					// if our hook is not the first in the linked list
					if (PrevHookPtr
						// Remove our hook from the linked list :
						// the next hook of PrevHookPtr becomes the next hook of HookPtr...
						// but only if the next hook follows Kevin's convention !
						&& (*(long*)(*(char**)(HookPtr+12) - 16) == EVHK_TAG_HEXA))
							*(char**)(PrevHookPtr+12) = *(char**)(HookPtr+12);
					else
						// Our hook is the first in the linked list :
						// EV_hook becomes the hook after our hook in the list
						EV_hook = (EVENT_HANDLER)*(char**)(HookPtr+12);
					// Free the block where the TSR was.
					// (don't forget to make HookPtr leave the ghost space)
					HeapFree (HeapPtrToHandle ((void*)((long)HookPtr ^ 0x40000)));
					fatal (PRGM_NAME" uninstalled.");
				}     // if(!memcmp(HookPtr, ...

				PrevHookPtr = HookPtr;     // HookPtr is now the previous hook pointer
				//Get the address of the entry point of the next hook
				HookPtr = *(char**)(HookPtr+12);
			}
	}     // if(EV_hook)
	

	{
		// search the file containing the icons & descriptions
		SYM_ENTRY *SymPtr;
		if(!(SymPtr = FindFile (ICO_FILENAME, ICO_TYPE, FALSE, FALSE, NULL, NULL)))
			fatal ("Error : '"ICO_FILENAME"' not found.");
		ICO_handle = SymPtr->handle;
	}

	// install h220xTSR
	if(!h220xTSR())
		fatal("Error : h220xTSR installation failed.");


	//=====================================================================
	// Now we create all the ACB and the object frames for the programs
	// of the calc referenced in the file containing the icons and the
	// descriptions, and for the text files if the viewer was found.
	//=====================================================================

	{
		char *FirstICO;
	
		// to keep ICO_Ptr valid even if we create blocks of memory, causing a
		// Heap Compression
		// don't forget to unlock it !
		if ((FirstICO = HLock (ICO_handle)) == NULL)
			fatal ("Not enough memory.");
		// jump the size of the file
		FirstICO += 2;
		// search for the text viewer
		TextViewerFound = FindTxtViewer();
		// get the list of all the supported files of the calc
		if (!(FilesList_handle = GetSupportedFilesList (TextViewerFound, &FilesNumber))) {
			HeapUnlock (ICO_handle);
			fatal ("Not enough memory.");
		}
		FilesListPtr_top = HeapDeref (FilesList_handle);
		
		if (*(((unsigned long*)FirstICO)++) != ICO_MAGIC) {     // not the right file
			HeapFree (FilesList_handle);
			HeapUnlock (ICO_handle);
			fatal ("Error : '"ICO_FILENAME"' not found.");
		}
		
		if (!CreateACBs (FirstICO, FilesListPtr_top, FilesNumber)) {
			// not enough memory
			HeapUnlock (ICO_handle);
			FreeCreatedACB_and_Frames();
			HeapFree (FilesList_handle);
			fatal ("Not enough memory.");
		}
		
		HeapUnlock (ICO_handle);     // we won't use it any more
	}     // create_ACBsAndObjectFrames block


	//=====================================================================
  // We create an ACB and an object frame for the fake app 'Refresh
  // Desktop'.
  // Its name is NULL (we do not need it, low memory will be read, it is
  // not really a problem).
  // the last argument but one is a pointer to the variable where the
  // AppID will be stored (it will be used when the backup of the folders
  // is restored).
  // The last argument indicate that we are creating an ACB for 'Refresh
  // Desktop.
  //=====================================================================
	if (!CreateOneACB (FALSE, &RefreshDesktopData, sizeof (RefreshDesktopData), \
	    NULL, GetLastACB(),	&RefreshDT_AppID, TRUE)) {
		FreeCreatedACB_and_Frames();
		HeapFree (FilesList_handle);
		fatal ("Not enough memory.");		
	}

	
	//=====================================================================
	// Now install the event hook
	//=====================================================================
	
	// initialize the variable for the header of the event hook
	OldEvHook = EV_hook;
	// install the event hook (with HeapAllocHigh(), NOT HeapAllocPtr(), too
	// allow the uninstallation with UNINEVHK (even if the user shouldn't use
	// this program to uninstall us, but I don't want UNINEVHK to threw an
	// address error or anything else)
	{
		HANDLE block_handle = HeapAllocHigh (&EndNewEvHook-&NewEvHook);
		if (!block_handle)
			fatal ("Not enough memory.");
		NewEvHookPtr = HeapDeref (block_handle);
	}
	memcpy (NewEvHookPtr, &NewEvHook, &EndNewEvHook - &NewEvHook);	
	// Don't forget that the entry point is after the header -> + 16
	// and don't forget to make it called in the ghost space
	EV_hook = (EVENT_HANDLER)(NewEvHookPtr + 0x40000 + 16);

	
	//=====================================================================
	// Now we restore the backup of the Desktop folders structure, if this
	// backup is found.
	//=====================================================================
	
	{
		SYM_ENTRY *SymPtr;
		if ((SymPtr = FindFile (BCK_FILENAME, BACKUP_TYPE, FALSE, FALSE, NULL, NULL))) {
			char *Bck_Ptr, *DTlist_Ptr, *DT_list_begin_Ptr;
			HANDLE DTlist_handle;
			unsigned short OldListSize, NewListSize, FilesToAddNum, i;
			char *folders_Ptr = DTfolders_Ptr;
			BckFolderStruct BckSt;            // these structures are unused,
			DT_FolderStruct FoldSt;           // we are just using them for
			FileOfFilesListStruct FileOfList; // sizeof()

			Bck_Ptr = HeapDeref (SymPtr->handle);
			Bck_Ptr += 2;     // jump the size of the file
			if (*((long*)Bck_Ptr)++ == BACKUP_MAGIC) {
				// It's a valid Backup file. Else consider there is no backup

				// folders_Ptr += sizeof (FoldSt.AppsListHandle), because we will be
				// pointing to the files list handle at the end of each loop
				for (i = DESKTOP_FOLDERS_NUM; i; i--, folders_Ptr += sizeof (FoldSt.AppsListHandle)) {
					// copy the name of the folder from the backup to the Desktop App
					// internal variable
					memcpy (folders_Ptr, Bck_Ptr, sizeof (FoldSt.folder_name));
					// Get the handle of the list of the files of this folder
					folders_Ptr += sizeof (FoldSt.folder_name) + sizeof (FoldSt.must_be_null);
					DTlist_handle = *(HANDLE*)folders_Ptr;
					Bck_Ptr += sizeof (BckSt.folder_name);     // jumps the name
					// get the number of prgms and apps to add to the list
					FilesToAddNum = *((unsigned short*)Bck_Ptr)++;
					if (!FilesToAddNum)     // no files to add
						continue;     // Bck_Ptr is pointing to the name of the next folder ->ok
				
					// if the list does not exists, creates it : 2 bytes (as if they
					// were the 2 null bytes ending the list : all the operations
					// done below won't mess up)
					if (!DTlist_handle) {
						DTlist_handle = HeapAlloc (sizeof (short));
						OldListSize = sizeof (short);     // and the size is 2
					}
					else
						// get the current size of the list
						OldListSize = GetAppsListSize (HeapDeref (DTlist_handle));			
					// Enlarge the list enough to store the ACB handles of all the
					// programs, the text files and the flash apps that should be
					// stored (but actually some programsn text files and apps won't
					// be added)
					// (the << 1 is because each AppID uses 2 bytes)
					// Note that we do not check if there is enough free memory.
					// If the block cannot be reallocated, the calc is really in a bad state.
					// The Desktop App won't be able to be run anyway.
					HeapRealloc (DTlist_handle, NewListSize = OldListSize + (FilesToAddNum << 1));
					DTlist_Ptr = DT_list_begin_Ptr = HeapDeref (DTlist_handle);
					// We will write on the null word ending the list -> -2
					OldListSize -= sizeof (short);
					// jump the AppID of the apps already in the folder
					DTlist_Ptr += OldListSize;
					// OldListSize will now be the number of apps in the folder before
					// having added any program or flash app
					OldListSize >>= 1;
	
					// Add each prgm and app
					// (Bck_Ptr += sizeof (BckSt.files[0].file_name to jump the name of the prgm or the app)
					for (; FilesToAddNum; FilesToAddNum--, Bck_Ptr += sizeof (BckSt.files[0].file_name)) {
						// the type of the file
						BCK_TAG file_type = *((BCK_TAG*)Bck_Ptr)++;
					
						if (file_type == BCK_FLASH_APP) {     // it's a flash app
							char buffer[9];
							AppID App_handle;
							unsigned short j;
							AppID *ScanPtr;
							
							// copy the name
							memcpy (buffer, Bck_Ptr, sizeof (BckSt.files[0].file_name));
							// null-terminated string
							buffer[sizeof (BckSt.files[0].file_name)] = 0;
							// is this flash app still installed on the calc ?
							App_handle = custom_EV_getAppID (buffer);
							if (!App_handle) {     // no
								NewListSize -= sizeof (AppID); // the list will contain 1 file less than expected
								continue;     // deal with the next file to add to this folder
							}
							// search if the AppID is already in the list
							// (the app would be already in the folder)
							for (j = OldListSize, ScanPtr = (AppID*)DT_list_begin_Ptr; j; j--) {
								if (*ScanPtr++ == App_handle)
									break;
							}
							if (*(ScanPtr-1) == App_handle) {     // we have broken
								NewListSize -= sizeof (AppID); // the list will contain 1 file less than expected	
								continue;     // don't copy this App in the folder
							}
							// else add the AppID at the end of the list
							*((AppID*)DTlist_Ptr)++ = App_handle;
							
						}    // if (is_flash_app)
						
						else if (file_type == BCK_PRGM || file_type == BCK_TEXT) {
							// it's an ASM/PPG/BASIC prgm or a TEXT file
							unsigned short i;
							char *FilesListPtr = FilesListPtr_top;
							
							// First we search for the name of the program in the list
							// we have created (Bck_Ptr points to the name in the backup file)
							short found = 0;     // = 0 if found (initialization to avoid a warning)
							for (i = FilesNumber; i; i--) {
								if (!(found = memcmp (FilesListPtr, Bck_Ptr, sizeof (BckSt.folder_name))))
									break;
								FilesListPtr += sizeof (FileOfList);     // next name in the list
							}

							// the AppID of the fake ACB for this file is just after the
							// name in the files list.
							// We add this AppID at the end of the list of the files
							// of the current folder, if it is not FALSE
							// (FileOfList.TwoBytes.Created) (-> the icon has been found
							// in the datafile).
							if (!found && *(BOOL*)(FilesListPtr + sizeof(FileOfList.name))) {
								// for TEXT files, the viewer must have been found, and the file
								// must really be a text file
								if (file_type == BCK_TEXT)
									if (!TextViewerFound \
									 || (((ACB*)HeapDeref (*(AppID*)(FilesListPtr + sizeof (FileOfList.name))))) \
									 ->publicstorage != ACB_TXT_MAGIC) {
										// then the list will contain 1 file less than expected
										NewListSize -= sizeof (AppID);
										continue;     // deal with the next file to restore in this folder
									}
								// else it's ok for text files and programs
								*((AppID*)DTlist_Ptr)++ = *(AppID*)(FilesListPtr + sizeof (FileOfList.name));
							}     // if (!found && *(BOOL*)
							else     // the file is not in the list, or it has no ACB
								NewListSize -= sizeof (AppID); // the list will contain 1 file less than expected

						}     // else if(<ASM/PPG/TI-BASIC>)
						
						else     // it's the fake app 'Refresh Desktop'
						  // it's AppID is RefreshDT_AppID, copy it in the Apps list
							*((AppID*)DTlist_Ptr)++ = RefreshDT_AppID;
						
					}     // for (; FilesToAddNum; ...
					
					// We have now restored the files of one folder.
					// the list must end with a null word
					*(short*)DTlist_Ptr = 0;
					// Now if the new list of the files is empty, we can delete it.
					if (NewListSize == sizeof (short))
						HeapFree (DTlist_handle);
					else {
						// Perhaps the list has not the size expected (some files have not
						// been added to the list). NewListSize is its real size.
						HeapRealloc (DTlist_handle, NewListSize);
						// After the list of the folders names of the Desktop, we must
						// refresh the handle of the list of the files that is now invalid.
						*(HANDLE*)folders_Ptr = DTlist_handle;
					}     // else

				}     // for (i = DESKTOP_FOLDERS_NUM; ...
			}     // if (*((long*)Bck_Ptr)++ == BACKUP_MAGIC)
		}     // if (SYM_ENTRY *SymPtr = ...
	}     // RestoreFoldersBackup block

	HeapFree (FilesList_handle);     // we do not need the list any more
	
	// Set Mode/Apps Desktop to ON
	MO_currentOptions();
	MO_option[MO_OPT_DESKTOP] = D_DESKTOP_ON;
	MO_digestOptions (H_NULL);
	ST_helpMsg (PRGM_NAME" "PRGM_VERSION" installed !");

} // _main()
